package org.photon.core.login.net;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.google.inject.Inject;
import com.google.inject.Injector;
import com.typesafe.config.Config;
import org.apache.mina.core.service.IoAcceptor;
import org.apache.mina.core.service.IoHandler;
import org.apache.mina.core.session.IdleStatus;
import org.apache.mina.core.session.IoSession;
import org.apache.mina.filter.codec.ProtocolCodecFactory;
import org.apache.mina.filter.codec.ProtocolCodecFilter;
import org.apache.mina.transport.socket.nio.NioSocketAcceptor;
import org.joda.time.Instant;
import org.photon.core.login.net.event.ClientStateUpdatedEvent;
import org.photon.core.login.net.event.HandleClientMessageEvent;
import org.photon.core.protocol.client.ClientState;
import org.photon.shared.event.EventDispatcher;
import org.photon.shared.event.EventDispatcherStrategy;
import org.photon.shared.event.SimpleEventDispatcher;
import org.photon.shared.kernel.AbstractService;
import org.photon.shared.kernel.State;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.util.Set;

import static com.google.common.base.Throwables.propagate;

/**
 * @author Blackrush
 */
public final class DefaultLoginNetworkService extends AbstractService implements LoginNetworkService, IoHandler {
    private static final Logger log = LoggerFactory.getLogger(DefaultLoginNetworkService.class);

    private final EventDispatcher dispatcher;
    private final IoAcceptor acceptor;
    private final Injector inject;
    private final Set<DefaultLoginClient> clients = Sets.newHashSet();

    @Inject
    public DefaultLoginNetworkService(Injector inject, EventDispatcherStrategy strategy, Config config, ProtocolCodecFactory codecFactory) {
        super(State.LOADED);

        this.inject = inject;

        this.dispatcher = SimpleEventDispatcher.create(strategy);

        this.acceptor = new NioSocketAcceptor();
        this.acceptor.setDefaultLocalAddress(getladdr(config.getString("photon.login.network.laddr")));
        this.acceptor.getFilterChain().addLast("codec", new ProtocolCodecFilter(codecFactory));
        this.acceptor.setHandler(this);
    }

    @Inject
    public void registerHandlers(Set<LoginNetworkHandler> handlers) {
        for (LoginNetworkHandler handler : handlers) {
            dispatcher.subscribe(handler);
        }
    }

    static SocketAddress getladdr(String value) {
        String[] laddr = value.split(":", 2);
        return new InetSocketAddress(laddr[0], Integer.parseInt(laddr[1]));
    }

    @Override
    public EventDispatcher getDispatcher() {
        return dispatcher;
    }

    @Override
    public Instant getStartInstant() {
        return getState() == State.STARTED ?
                new Instant(acceptor.getActivationTime()) :
                null;
    }

    @Override
    public Set<LoginClient> getConnectedClients() {
        return ImmutableSet.<LoginClient>copyOf(clients);
    }

    @Override
    public int getCurrentConnectedClients() {
        return acceptor.getManagedSessionCount();
    }

    @Override
    public int getMaxConnectedClients() {
        return acceptor.getStatistics().getLargestManagedSessionCount();
    }

    @Override
    protected void setUp() {
        try {
            acceptor.bind();
            log.debug("bound to {}", acceptor.getDefaultLocalAddress());
        } catch (IOException e) {
            throw propagate(e);
        }
    }

    @Override
    protected void tearDown() {
        acceptor.dispose(true);
        acceptor.unbind();
    }

    private DefaultLoginClient createClient(IoSession session) {
        DefaultLoginClient client = new DefaultLoginClient(session);
        client.setState(ClientState.CHECK_VERSION);

        inject.injectMembers(client);
        clients.add(client);

        return client;
    }

    @Override
    public void sessionCreated(IoSession session) throws Exception {
        session.setAttribute(this, createClient(session));
    }

    @Override
    public void sessionOpened(IoSession session) throws Exception {
        log.debug("session {} opened", session.getRemoteAddress());

        DefaultLoginClient client = (DefaultLoginClient) session.getAttribute(this);
        dispatcher.dispatch(new ClientStateUpdatedEvent(client, true));
    }

    @Override
    public void sessionClosed(IoSession session) throws Exception {
        log.debug("session {} closed", session.getRemoteAddress());

        DefaultLoginClient client = (DefaultLoginClient) session.removeAttribute(this);
        dispatcher.dispatch(new ClientStateUpdatedEvent(client, false));
    }

    @Override
    public void exceptionCaught(IoSession session, Throwable cause) throws Exception {
        log.error("unhandled exception (remoteaddr=" + session.getRemoteAddress() + ")", cause);
    }

    @Override
    public void messageReceived(IoSession session, Object o) throws Exception {
        log.debug("input message '{}' from {}", o, session.getRemoteAddress());

        DefaultLoginClient client = (DefaultLoginClient) session.getAttribute(this);
        dispatcher.dispatch(new HandleClientMessageEvent(client, o));
    }

    @Override
    public void messageSent(IoSession session, Object o) throws Exception {
        log.debug("output message '{}' to {}", o, session.getRemoteAddress());
    }

    @Override
    public void sessionIdle(IoSession session, IdleStatus status) throws Exception { }
}
