package org.photon.core.login.net.game;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.google.inject.Inject;
import com.typesafe.config.Config;
import org.apache.mina.core.future.IoFutureListener;
import org.apache.mina.core.service.IoAcceptor;
import org.apache.mina.core.service.IoHandler;
import org.apache.mina.core.session.IdleStatus;
import org.apache.mina.core.session.IoSession;
import org.apache.mina.filter.codec.ProtocolCodecFilter;
import org.apache.mina.filter.codec.serialization.ObjectSerializationCodecFactory;
import org.apache.mina.transport.socket.nio.NioSocketAcceptor;
import org.photon.core.protocol.server.AuthenticationErrorMessage;
import org.photon.core.protocol.server.AuthenticationMessage;
import org.photon.core.protocol.server.AuthenticationSuccessMessage;
import org.photon.core.protocol.server.HelloConnectMessage;
import org.photon.shared.event.EventDispatcher;
import org.photon.shared.event.EventDispatcherStrategy;
import org.photon.shared.event.SimpleEventDispatcher;
import org.photon.shared.kernel.AbstractService;
import org.photon.shared.kernel.State;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.util.Map;
import java.util.Set;

import static com.google.common.base.Throwables.propagate;

/**
 * @author Blackrush
 */
public final class DefaultLoginGameManager extends AbstractService implements LoginGameManager, IoHandler {
    private static final String SERVICE_ATTR = DefaultLoginGameManager.class.getName() + ".service_attr";

    private final Map<Integer, DefaultLoginGameService> services = Maps.newHashMap();
    private final IoAcceptor acceptor;
    private final EventDispatcher dispatcher;

    @Inject
    public DefaultLoginGameManager(Config config, EventDispatcherStrategy dispatcherStrategy) {
        super(State.LOADED);

        this.acceptor = new NioSocketAcceptor();
        this.acceptor.setDefaultLocalAddress(getladdr(config.getString("photon.login.network.game.laddr")));
        this.acceptor.getFilterChain().addLast("codec", new ProtocolCodecFilter(new ObjectSerializationCodecFactory()));
        this.acceptor.setHandler(this);

        this.dispatcher = SimpleEventDispatcher.create(dispatcherStrategy);

        for (Integer serverId : config.getIntList("photon.login.network.game.servers")) {
            services.put(serverId, new DefaultLoginGameService(serverId, this));
        }
    }

    static SocketAddress getladdr(String value) {
        String[] laddr = value.split(":", 2);
        return new InetSocketAddress(laddr[0], Integer.parseInt(laddr[1]));
    }

    @Override
    public EventDispatcher getDispatcher() {
        return dispatcher;
    }

    @Override
    public Set<LoginGameService> getServices() {
        return ImmutableSet.<LoginGameService>copyOf(services.values());
    }

    @Override
    protected void setUp() {
        try {
            acceptor.bind();
        } catch (IOException e) {
            throw propagate(e);
        }
    }

    @Override
    protected void tearDown() {
        acceptor.dispose(true);
        acceptor.unbind();
    }

    @Override
    public void exceptionCaught(IoSession session, Throwable cause) throws Exception {

    }

    @Override
    public void sessionCreated(IoSession session) throws Exception {

    }

    @Override
    public void sessionOpened(IoSession session) throws Exception {
        session.write(new HelloConnectMessage());
    }

    @Override
    public void sessionClosed(IoSession session) throws Exception {
        DefaultLoginGameService service = (DefaultLoginGameService) session.getAttribute(SERVICE_ATTR);
        if (service == null) {
            return;
        }

        service.desync();
    }

    @Override
    public void messageReceived(IoSession session, Object o) throws Exception {
        DefaultLoginGameService service = (DefaultLoginGameService) session.getAttribute(SERVICE_ATTR);
        if (service != null) {
            service.onMessage(o);
        } else if (o instanceof AuthenticationMessage) {
            AuthenticationMessage message = (AuthenticationMessage) o;

            service = services.get(message.getId());
            if (service == null) {
                session.write(new AuthenticationErrorMessage(AuthenticationErrorMessage.Reason.UNKNOWN_SERVER))
                        .addListener(IoFutureListener.CLOSE);
            } else if (service.isSync()) {
                session.write(new AuthenticationErrorMessage(AuthenticationErrorMessage.Reason.ALREADY_SYNC))
                       .addListener(IoFutureListener.CLOSE);
            } else {
                session.setAttribute(SERVICE_ATTR, service);
                service.sync(session, message.getLaddr());
                session.write(new AuthenticationSuccessMessage());
            }
        } else {
            throw new IllegalStateException("input message's session musts be synchronized or musts try to authenticate");
        }
    }

    @Override
    public void messageSent(IoSession session, Object message) throws Exception {

    }

    @Override public void sessionIdle(IoSession session, IdleStatus status) throws Exception { }
}
