package org.photon.core.login.net.handler;

import com.google.common.collect.Collections2;
import com.google.inject.Inject;
import org.photon.core.login.db.AuthenticationException;
import org.photon.core.login.db.User;
import org.photon.core.login.db.UserRepository;
import org.photon.core.login.net.LoginClient;
import org.photon.core.login.net.LoginNetworkHandler;
import org.photon.core.login.net.LoginNetworkService;
import org.photon.core.login.net.event.ClientStateUpdatedEvent;
import org.photon.core.login.net.event.HandleClientMessageEvent;
import org.photon.core.login.net.game.LoginGameManager;
import org.photon.core.login.net.game.LoginGameService;
import org.photon.core.login.net.game.event.GameServiceStateUpdatedEvent;
import org.photon.core.protocol.client.ClientState;
import org.photon.core.protocol.client.enums.ClientAuthErrorReason;
import org.photon.core.protocol.client.messages.*;
import org.photon.shared.event.Subscribe;
import org.photon.shared.inject.InjectConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Blackrush
 */
public final class DefaultLoginNetworkHandler implements LoginNetworkHandler {
    private static final Logger log = LoggerFactory.getLogger(DefaultLoginNetworkHandler.class);

    @InjectConfig("photon.login.client_version") String clientVersion;
    @InjectConfig("photon.login.community") int community;
    @Inject UserRepository users;
    @Inject LoginGameManager game;
    @Inject LoginNetworkService network;

    @Inject
    public DefaultLoginNetworkHandler() {
        game.getDispatcher().subscribe(this);
    }

    @Subscribe
    public void onClientUpdate(ClientStateUpdatedEvent event) {
        if (event.isConnection()) {
            event.getClient().write(new HelloConnectMessage(event.getClient().getKey()));
        }
    }

    @Subscribe
    public void onGameServiceStateUpdated(GameServiceStateUpdatedEvent event) {
        for (LoginClient client : network.getConnectedClients()) {
            if (client.getState() != ClientState.OTHER) continue; // make sure he is on server selection page
            client.write(new GameServerMessage(event.getService().toGameServerType()));
        }
    }

    @Subscribe(HandleClientMessageEvent.class)
    public void onClientVersion(LoginClient client, ClientVersionMessage message) {
        if (!clientVersion.equals(message.getVersion())) {
            client.kick();
        } else {
            client.setState(ClientState.AUTHENTICATION);
        }
    }

    @Subscribe(HandleClientMessageEvent.class)
    public void onUserCredentials(LoginClient client, UserCredentialsMessage message) {
        try {
            User user = users.authenticate(message.getUsername(), message.getPasshash(), client.getKey());
            client
                    .setUser(user)
                    .setState(ClientState.OTHER)
                    .write(new UserNicknameMessage(user.getNickname()))
                    .write(new UserCommunityMessage(community))
                    .write(new UserQuestionMessage(user.getSecretQuestion()))
                    .write(new GameServersMessage(Collections2.transform(game.getServices(), LoginGameService.TO_GAME_SERVER_TYPE)))
                    .write(new ClientAuthMessage(false)) // TODO user's rights
            ;
        } catch (AuthenticationException.InvalidCredentials e) {
            client.write(new ClientAuthErrorMessage(ClientAuthErrorReason.INVALID_CREDENTIALS)).kick();
        } catch (AuthenticationException.Banned e) {
            client.write(new ClientAuthErrorMessage(ClientAuthErrorReason.BANNED)).kick();
        }
    }

    @Subscribe(HandleClientMessageEvent.class)
    public void onPlayersListReq(LoginClient client, PlayersListReqMessage message) {

    }

    @Subscribe(HandleClientMessageEvent.class)
    public void onGameServerSelection(LoginClient client, GameServerSelectionMessage message) {

    }

    @Subscribe(HandleClientMessageEvent.class)
    public void onQueueReq(LoginClient client, QueueReqMessage message) {

    }
}
