package org.photon.core.protocol.client;

import org.apache.mina.core.session.IoSession;
import org.apache.mina.filter.codec.*;
import org.apache.mina.filter.codec.textline.LineDelimiter;
import org.apache.mina.filter.codec.textline.TextLineDecoder;
import org.apache.mina.filter.codec.textline.TextLineEncoder;
import org.photon.core.protocol.client.messages.ClientVersionMessage;
import org.photon.core.protocol.client.messages.UserCredentialsMessage;
import org.photon.shared.collect.StringTree;

import java.nio.charset.Charset;

/**
 * @author Blackrush
 */
public class PhotonMinaCodecFactory implements ProtocolCodecFactory {
    public static final Charset CHARSET = Charset.forName("UTF-8");
    public static final LineDelimiter ENCODING_DELIMITER = LineDelimiter.NUL,
            DECODING_DELIMITER = new LineDelimiter("\n\0");

    public static final String CLIENT_STATE_ATTR = PhotonMinaCodecFactory.class.getName() + ".client_state";

    private final StringTree<MessageFactory> factories;

    public PhotonMinaCodecFactory(StringTree<MessageFactory> factories) {
        this.factories = factories;
    }

    @Override
    public ProtocolEncoder getEncoder(IoSession session) throws Exception {
        return new Encoder();
    }

    @Override
    public ProtocolDecoder getDecoder(IoSession session) throws Exception {
        return new Decoder();
    }

    public class Encoder extends TextLineEncoder {
        public Encoder() {
            super(CHARSET, ENCODING_DELIMITER);
        }

        @Override
        public void encode(IoSession session, Object message, ProtocolEncoderOutput out) throws Exception {
            if (message instanceof Message) {
                StringBuilder builder = new StringBuilder();
                ((Message) message).serialize(builder);
                message = builder.toString();
            }

            super.encode(session, message, out);
        }
    }

    public class Decoder extends TextLineDecoder {
        public Decoder() {
            super(CHARSET, DECODING_DELIMITER);
        }

        @Override
        protected void writeText(IoSession session, String text, ProtocolDecoderOutput out) {
            Message message = null;

            switch ((ClientState) session.getAttribute(CLIENT_STATE_ATTR, ClientState.OTHER)) {
            case CHECK_VERSION:
                message = new ClientVersionMessage();
                break;
            case AUTHENTICATION:
                message = new UserCredentialsMessage();
                break;
            default:
                MessageFactory factory = factories.get(text);
                if (factory != null) {
                    message = factory.buildMessage();
                }
                break;
            }

            if (message != null) {
                message.deserialize(text);
                out.write(message);
            } else {
                out.write(text);
            }
        }
    }
}
