package org.photon.shared.kernel;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;

import java.io.File;
import java.io.FilenameFilter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Collection;
import java.util.Set;
import java.util.jar.JarFile;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Throwables.propagate;

/**
 * @author Blackrush
 */
public class FileBundleLoader extends BundleLoader {
    public static final String BUNDLE_CLASS_ATTR = "Bundle-Class";
    private final File root, libRoot;
    private final ClassLoader masterLoader;

    public FileBundleLoader(File root, File libRoot, ClassLoader masterLoader) {
        this.root = checkNotNull(root);
        this.libRoot = checkNotNull(libRoot);
        this.masterLoader = checkNotNull(masterLoader);
    }

    private static <T> Collection<T> concat(Collection<T>... collections) {
        ImmutableSet.Builder<T> builder = ImmutableSet.builder();
        for (Collection<T> collection : collections) {
            builder.addAll(collection);
        }
        return builder.build();
    }

    @Override
    public Set<Bundle> load() {
        Set<File> files = getFileBundles();
        URLClassLoader loader = new URLClassLoader(getUrls(concat(files, getFileLibs())), masterLoader);

        Set<Bundle> bundles = Sets.newHashSet();

        try {
            for (File file : files) {
                JarFile jarfile = new JarFile(file);

                String bundleClassName = jarfile.getManifest().getMainAttributes().getValue(BUNDLE_CLASS_ATTR);

                if (bundleClassName == null) {
                    throw new IllegalStateException(file.getAbsolutePath() + " must have a Bundle-Class attribute in its manifest file");
                }

                Class<?> bundleClass = loader.loadClass(bundleClassName);

                if (!Bundle.class.isAssignableFrom(bundleClass)) {
                    throw new IllegalStateException(bundleClassName + " must be a " + Bundle.class.getName());
                }

                bundles.add((Bundle) bundleClass.newInstance());
            }
        } catch (Exception e) {
            throw propagate(e);
        }

        return bundles;
    }

    private Set<File> getFileBundles() {
        Set<File> result = Sets.newHashSet();
        getFileBundles(result, root);
        return result;
    }

    private Set<File> getFileLibs() {
        return Sets.newHashSet(libRoot.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.endsWith(".jar");
            }
        }));
    }

    private URL[] getUrls(Collection<File> files) {
        URL[] urls = new URL[files.size()];

        int i = 0;
        for (File file : files) {
            try {
                urls[i++] = file.toURI().toURL();
            } catch (MalformedURLException e) {
                throw propagate(e);
            }
        }

        return urls;
    }

    private void getFileBundles(Set<File> result, File file) {
        if (file.isDirectory()) {
            for (File child : file.listFiles()) {
                getFileBundles(result, child);
            }
        } else {
            result.add(file);
        }
    }
}
