package org.photon.shared.kernel;

import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Module;
import com.google.inject.Stage;
import com.typesafe.config.Config;
import org.photon.shared.event.EventDispatcher;
import org.photon.shared.event.EventDispatcherStrategy;
import org.photon.shared.event.ThreadedEventDispatcher;
import org.photon.shared.inject.ConfigModule;
import org.photon.shared.kernel.event.ChangeStateEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Iterator;
import java.util.List;
import java.util.Set;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Throwables.propagate;

/**
 * @author Blackrush
 */
public class SimpleBundleRegistry extends BundleRegistry {
    private static final Logger log = LoggerFactory.getLogger(SimpleBundleRegistry.class);

    private final List<ServiceInterface> services = Lists.newArrayList();
    private final BundleLoader loader;
    private State state = State.NONE;
    private Set<Bundle> bundles;

    private final EventDispatcher dispatcher;
    private final Stage stage;
    private final Set<Module> initialModules;
    private Config config;
    private Injector injector;

    /**
     * @param stage can be null
     * @param config must not be null
     * @param loader must not be null
     * @param initialModules can be null
     * @param dispatcherStrategy see {@link EventDispatcherStrategy} constructor
     */
    public SimpleBundleRegistry(Stage stage, Config config, BundleLoader loader, Set<Module> initialModules, EventDispatcherStrategy dispatcherStrategy) {
        this.stage = stage != null ? stage : Stage.DEVELOPMENT;
        this.config = checkNotNull(config);
        this.loader = checkNotNull(loader);
        this.initialModules = initialModules != null ? initialModules : Sets.<Module>newHashSet();
        this.dispatcher = ThreadedEventDispatcher.create(dispatcherStrategy);
    }

    @Override
    public Injector getInjector() {
        return injector;
    }

    @Override
    public EventDispatcher getDispatcher() {
        return dispatcher;
    }

    @Override
    public State getState() {
        return state;
    }

    protected void setState(State state) {
        this.state = state;
        this.dispatcher.dispatch(ChangeStateEvent.create(this, state));
    }

    @Override
    public void load() {
        if (state != State.NONE) {
            throw new IllegalStateException("this registry is already loaded");
        }

        bundles = loader.load();

        loadBundles();
        injector = Guice.createInjector(stage, initialModules);

        loadServices();

        for (Bundle bundle : bundles) {
            bundle.load(this);
        }

        setState(State.LOADED);
    }

    private void loadBundles() {
        for (Bundle bundle : bundles) {
            for (Class<? extends Module> moduleClass : bundle.getModules()) {
                try {
                    Module module = moduleClass.newInstance();
                    initialModules.add(module);
                } catch (Exception e) {
                    throw propagate(e);
                }
            }

            config = config.withFallback(bundle.getDefaultParameters());

            bundle.load(this);
            log.debug("bundle '{}' loaded", bundle.getIdentifier());
        }

        initialModules.add(new BindThisRegistryModule());
        initialModules.add(ConfigModule.of(config));
    }

    private void loadServices() {
        for (Bundle bundle : bundles) {
            for (Class<? extends ServiceInterface> serviceClass : bundle.getServices()) {
                ServiceInterface service = injector.getInstance(serviceClass);
                services.add(service);

                log.debug("{} loaded", service.getClass().getName());
            }
        }
    }

    @Override
    public Bundle get(String name) {
        for (Bundle bundle : bundles) {
            if (bundle.getIdentifier().equals(name)) {
                return bundle;
            }
        }

        return null;
    }

    @Override
    public Config getConfig() {
        return config;
    }

    @Override
    public Iterator<Bundle> iterator() {
        return bundles.iterator();
    }

    @Override
    public void startUp() {
        if (state != State.LOADED) {
            throw new IllegalStateException("you must load this registry before starting it");
        }

        //noinspection ForLoopReplaceableByForEach
        for (int i = 0; i < services.size(); i++) {
            ServiceInterface service = services.get(i);
            service.startUp();
            log.debug("{} started", service.getClass().getName());
        }

        setState(State.STARTED);
    }

    @Override
    public void shutDown() {
        if (state != State.STARTED) {
            throw new IllegalStateException("this registry must be started before trying to stop it");
        }

        for (int i = services.size() - 1; i >= 0; i--) {
            ServiceInterface service = services.get(i);
            service.shutDown();
            log.debug("{} stopped", service.getClass().getName());
        }

        setState(State.STOPPED);
    }
}
